#include		"Intersect.h"

void clIntersect::funcGetIntervalSphere( clSphere a, clVec2 axis, float & fMin, float & fMax )
{
	fMin = ( a.vPos - axis * a.fRadius ) & axis;
	fMax = ( a.vPos + axis * a.fRadius ) & axis;

	if( fMin > fMax )
	{
		float	fTemp = fMin;
		fMin = fMax;
		fMax = fTemp;
	}
}

void clIntersect::funcGetIntervalLine( clLine a, clVec2 axis, float & fMin, float & fMax )
{
	clVec2	v;

	v = a.vMid - a.vDir * a.fHalf;
	fMin = v & axis;
	v = a.vMid + a.vDir * a.fHalf;
	fMax = v & axis;

	if( fMax < fMin )
	{
		float	fTemp = fMax;
		fMax = fMin;
		fMin = fTemp;
	}
}

void clIntersect::funcGetIntervalAABB( clAABB a, clVec2 axis, float & fMin, float & fMax )
{
	fMin = ( a.vPos - a.vSiz ) & ( axis );
	fMax = ( a.vPos - a.vSiz ) & ( axis );

	float	fTemp = clVec2( a.vPos.x - a.vSiz.x, a.vPos.y + a.vSiz.y ) & ( axis );
	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
	fTemp = ( a.vPos + a.vSiz ) & ( axis );
	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
	fTemp = clVec2( a.vPos.x + a.vSiz.x, a.vPos.y - a.vSiz.y ) & ( axis );
	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
}

void clIntersect::funcGetIntervalOBB( clOBB a, clVec2 axis, float & fMin, float & fMax )
{
	clVec2	v[ 4 ];
	clVec2	r[ 4 ];

	r[ 0 ].VecSet( - a.vSiz.x, - a.vSiz.y );
	r[ 1 ].VecSet( - a.vSiz.x, + a.vSiz.y );
	r[ 2 ].VecSet( + a.vSiz.x, + a.vSiz.y );
	r[ 3 ].VecSet( + a.vSiz.x, - a.vSiz.y );

	r[ 0 ].VecRotate( a.fAngle );
	r[ 1 ].VecRotate( a.fAngle );
	r[ 2 ].VecRotate( a.fAngle );
	r[ 3 ].VecRotate( a.fAngle );

	v[ 0 ] = a.vPos + r[ 0 ];
	v[ 1 ] = a.vPos + r[ 1 ];
	v[ 2 ] = a.vPos + r[ 2 ];
	v[ 3 ] = a.vPos + r[ 3 ];

	fMin = v[ 0 ] & axis;
	fMax = v[ 0 ] & axis;

	float	fTemp = v[ 1 ] & axis;

	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}

	fTemp = v[ 2 ] & axis;

	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}

	fTemp = v[ 3 ] & axis;

	if( fMin > fTemp )
	{
		fMin = fTemp;
	}
	else
	{
		if( fMax < fTemp )
		{
			fMax = fTemp;
		}
	}
}

bool clIntersect::funcIntersectAABBxAABB( clAABB & a0, clAABB & a1 )
{
	clVec2	vDel = a1.vPos - a0.vPos;	//    

	//    vDel   X  Y <=     
	return ( ( fabs( vDel.x ) <= ( a0.vSiz.x + a1.vSiz.x ) ) && 
			 ( fabs( vDel.y ) <= ( a0.vSiz.y + a1.vSiz.y ) ) );
}

bool clIntersect::funcIntersectSpherexSphere( clSphere & a0, clSphere & a1 )
{
	float	fRadx2 = a0.fRadius + a1.fRadius;	//  
	float	fDist = ( a0.vPos - a1.vPos ) & ( a0.vPos - a1.vPos );	//  

	//           
	return ( fRadx2 * fRadx2 >= fDist );
}

bool clIntersect::funcIntersectAABBxSphere( clAABB & a0, clSphere & a1 )
{
	float		fDist = 0.0f;

	//        
	float	f = a0.vPos.x - a0.vSiz.x;
	if( a1.vPos.x < f )
	{
		f = a1.vPos.x - f;
		fDist += f * f;
	}
	else
	{
		f = a0.vPos.x + a0.vSiz.x;
		if( a1.vPos.x > f )
		{
			f = a1.vPos.x - f;
			fDist += f * f;
		}
	}
	f = a0.vPos.y - a0.vSiz.y;
	if( a1.vPos.y < f )
	{
		f = a1.vPos.y - f;
		fDist += f * f;
	}
	else
	{
		f = a0.vPos.y + a0.vSiz.y;
		if( a1.vPos.y > f )
		{
			f = a1.vPos.y - f;
			fDist += f * f;
		}
	}
	//          
	return ( fDist <= a1.fRadius * a1.fRadius );
}

bool clIntersect::funcIntersectSpherexLine( clSphere & a0, clLine & a1 )
{
	float	fsmin, fsmax;
	float	flmin, flmax;

	//    A   B
	//      3 
	// 1     A
	clVec2	vAxis = ( a0.vPos - ( a1.vMid - a1.vDir * a1.fHalf ) );
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, fsmin, fsmax );
	funcGetIntervalLine( a1, vAxis, flmin, flmax );
	if( ( fsmin > flmax ) || fsmax < flmin )
	{
		return false;
	}

	// 2     B
	vAxis = ( a0.vPos - ( a1.vMid + a1.vDir * a1.fHalf ) );
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, fsmin, fsmax );
	funcGetIntervalLine( a1, vAxis, flmin, flmax );
	if( ( fsmin > flmax ) || ( fsmax < flmin ) )
	{
		return false;
	}

	// 3           
	vAxis = a1.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, fsmin, fsmax );
	funcGetIntervalLine( a1, vAxis, flmin, flmax );
	if( ( fsmin > flmax ) || ( fsmax < flmin ) )
	{
		return false;
	}

	//    ,   
	return true;
}

bool clIntersect::funcIntersectSpherexOBB( clSphere & a0, clOBB & a1 )
{
	clSphere	sphere;
	clAABB		aabb;

	clVec2		vDel = a0.vPos - a1.vPos;	//    
	//     ,      OBB
	vDel.VecRotate( - a1.fAngle );
	sphere.vPos = a1.vPos + vDel;	//    
	sphere.fRadius = a0.fRadius;	// 
	//   AABB
	aabb.vPos = a1.vPos;
	aabb.vSiz = a1.vSiz;

	//    AABB x Sphere
	return ( funcIntersectAABBxSphere( aabb, sphere ) );
}

bool clIntersect::funcIntersectAABBxLine( clAABB & a0, clLine & a1 )
{
	float	fMin, fMax;
	clVec2	vDel = a0.vPos - a1.vMid;	//   

	//    +       ,   
	if( ( fabs( vDel.x ) > a0.vSiz.x + fabs( a1.vDir.x ) * a1.fHalf ) ||
		( fabs( vDel.y ) > a0.vSiz.y + fabs( a1.vDir.y ) * a1.fHalf ) )
	{
		return false;
	}

	//  AABB      
	clVec2	vNor = a1.vDir.VecNormal();
	funcGetIntervalAABB( a0, vNor, fMin, fMax );

	float	fDot = a1.vMid & ( vNor );

	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	return true;
}

bool clIntersect::funcIntersectOBBxLine( clOBB & a0, clLine & a1 )
{
	clAABB		aabb;
	clLine		line;

	aabb.vPos = a0.vPos;
	aabb.vSiz = a0.vSiz;

	//       OBB
	clVec2		vDel = a1.vMid - a0.vPos;
	vDel.VecRotate( - a0.fAngle );
	line.vMid = a0.vPos + vDel;
	
	//  
	vDel = a1.vDir;
	vDel.VecRotate( - a0.fAngle );
	line.vDir = vDel;

	//    
	line.fHalf = a1.fHalf;

	return ( funcIntersectAABBxLine( aabb, line ) );
}

bool clIntersect::funcIntersectLinexLine( clLine & a0, clLine & a1 )
{
	float	fMin, fMax;

	//     a1
	clVec2	vAxis = a1.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalLine( a0, vAxis, fMin, fMax );
	float	fDot = a1.vMid & vAxis;
	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	//     a0
	vAxis = a0.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalLine( a1, vAxis, fMin, fMax );
	fDot = a0.vMid & vAxis;
	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	return true;
}

bool clIntersect::funcIntersectAABBxOBB( clAABB & a0, clOBB & a1 )
{
	float		famin, famax;
	float		fbmin, fbmax;

	//     AABB,    1, 0  0, 1      
	clVec2	vAxis( 1.0f, 0.0f );
	funcGetIntervalAABB( a0, vAxis, famin, famax );
	funcGetIntervalOBB( a1, vAxis, fbmin, fbmax );

	if( ( famax < fbmin ) || ( famin > fbmax ) )
	{
		return false;
	}

	vAxis.VecSet( 0.0f, 1.0f );
	funcGetIntervalAABB( a0, vAxis, famin, famax );
	funcGetIntervalOBB( a1, vAxis, fbmin, fbmax );

	if( ( famax < fbmin ) || ( famin > fbmax ) )
	{
		return false;
	}

	//    OBB

	float		fCos = cos( a1.fAngle );
	float		fSin = sin( a1.fAngle );

	vAxis.VecSet( fCos, fSin );
	funcGetIntervalAABB( a0, vAxis, famin, famax );
	funcGetIntervalOBB( a1, vAxis, fbmin, fbmax );

	if( ( famax < fbmin ) || ( famin > fbmax ) )
	{
		return false;
	}

	vAxis.VecSet( - fSin, fCos );
	funcGetIntervalAABB( a0, vAxis, famin, famax );
	funcGetIntervalOBB( a1, vAxis, fbmin, fbmax );

	if( ( famax < fbmin ) || ( famin > fbmax ) )
	{
		return false;
	}

	return true;
}

bool clIntersect::funcIntersectOBBxOBB( clOBB & a0, clOBB & a1 )
{
	clVec2	vDel = a1.vPos - a0.vPos;

	// AABB   OBB
	clAABB	aabb;
	aabb.vPos = a0.vPos;
	aabb.vSiz = a0.vSiz;

	//   OBB    
	clOBB	obb;
	vDel.VecRotate( - a0.fAngle );
	obb.fAngle = a1.fAngle - a0.fAngle;
	obb.vPos = a0.vPos + vDel;
	obb.vSiz = a1.vSiz;

	return ( funcIntersectAABBxOBB( aabb, obb ) );
}